import asyncio
from datetime import datetime
from decimal import Decimal

import pytest
import sqlalchemy as sa

import aiochsa
from aiochsa import error_codes
from aiochsa.dialect import ClickhouseSaDialect


def pytest_collection_modifyitems(items):
    for item in items:
        if asyncio.iscoroutinefunction(item.function):
            item.add_marker('asyncio')


@pytest.fixture(scope='session')
def dsn(docker_services):
    docker_services.start('clickhouse-server')
    public_port = docker_services.wait_for_service('clickhouse-server', 8123)
    return f'clickhouse://{docker_services.docker_ip}:{public_port}'


@pytest.fixture
async def pool(dsn):
    async with aiochsa.create_pool(dsn) as pool:
        yield pool


@pytest.fixture
async def conn(dsn):
    async with aiochsa.connect(dsn) as conn:
        yield conn


@pytest.fixture
def recreate_table(conn):

    async def _recreate(table_name, create_ddl):
        try:
            await conn.execute(f'DROP TABLE {table_name}')
        except aiochsa.DBException as exc:
            if exc.code != error_codes.UNKNOWN_TABLE:
                raise

        await conn.execute(create_ddl)

    return _recreate


TEST1_CREATE_DDL = '''\
CREATE TABLE test1
(
    id UInt64,
    enum Enum8(''=0, 'ONE'=1, 'TWO'=2),
    name String,
    timestamp DateTime,
    amount Decimal64(9)
)
ENGINE = MergeTree()
ORDER BY id
'''


@pytest.fixture
async def table_test1(recreate_table):
    await recreate_table('test1', TEST1_CREATE_DDL)

    return sa.Table(
        'test1', sa.MetaData(),
        sa.Column('id', sa.Integer),
        sa.Column('enum', sa.Enum, default=''),
        sa.Column('name', sa.String),
        sa.Column('timestamp', sa.DateTime, default=datetime.utcnow),
        sa.Column('amount', sa.DECIMAL, default=Decimal(0)),
    )


TEST2_CREATE_DDL = '''\
CREATE TABLE test2
(
    title String,
    num UInt64
)
ENGINE = MergeTree()
ORDER BY num
'''


@pytest.fixture
async def table_test2(recreate_table):
    await recreate_table('test2', TEST2_CREATE_DDL)

    return sa.Table(
        'test2', sa.MetaData(),
        sa.Column('title', sa.String),
        sa.Column('num', sa.Integer),
    )


TEST3_CREATE_DDL = '''\
CREATE TABLE test3
(
    key UInt8,
    value UInt64
)
ENGINE = SummingMergeTree()
ORDER BY key
'''


@pytest.fixture
async def table_test3(recreate_table):
    await recreate_table('test3', TEST3_CREATE_DDL)

    return sa.Table(
        'test3', sa.MetaData(),
        sa.Column('key', sa.Integer),
        sa.Column('value', sa.Integer),
    )


TEST4_CREATE_DDL = '''\
CREATE TABLE test4
(
    key UInt8,
    value AggregateFunction(sum, UInt8)
)
ENGINE = AggregatingMergeTree()
ORDER BY key
'''


@pytest.fixture
async def table_test4(recreate_table):
    await recreate_table('test4', TEST4_CREATE_DDL)

    return sa.Table(
        'test4', sa.MetaData(),
        sa.Column('key', sa.Integer),
        sa.Column('value', sa.Integer),
    )


TABLE_FOR_TYPE_DDL_TEMPLATE = '''\
CREATE TABLE test_for_type
(
    value {type}
)
ENGINE = Log()
'''


@pytest.fixture
def recreate_table_for_type(recreate_table):

    async def _recreate(type_str):
        create_ddl = TABLE_FOR_TYPE_DDL_TEMPLATE.format(type=type_str)
        await recreate_table('test_for_type', create_ddl)
        return 'test_for_type'

    return _recreate


@pytest.fixture
def table_for_type(recreate_table_for_type):

    async def _create(sa_type):
        if isinstance(sa_type, type):
            sa_type = sa_type()
        type_str = sa_type.compile(dialect=ClickhouseSaDialect())

        table_name = await recreate_table_for_type(type_str)

        return sa.Table(
            table_name, sa.MetaData(),
            sa.Column('value', sa_type),
        )

    return _create



import re


# Based on `getExceptionMessage()`:
# https://github.com/yandex/ClickHouse/blob/master/dbms/src/Common/Exception.cpp#L261
_match_exc_message = re.compile(
    r'^Code: (?P<code>\d+), '
        r'e\.displayText\(\) = (?P<display_text>.+?)'
        r'(?:, Stack trace[^:]*:\s+(?P<stack_trace>.+))?$',
    re.M,
).match


class AiochsaException(Exception):
    """ Base class for aiochsa exceptions """


class ProtocolError(AiochsaException):
    """ Error communicating to Clickhouse server """


class DBException(AiochsaException):
    """ Error returned from Clickhouse database """

    def __init__(self, statement, code, display_text, stack_trace=None):
        super().__init__(statement, code, display_text, stack_trace)
        self.statement = statement
        self.code = code
        self.display_text = display_text
        self.stack_trace = stack_trace

    def __str__(self):
        statement = self.statement
        if len(statement) > 200:
            statement = statement[:200] + '...'
        return f'[Code={self.code}] {self.display_text}: {statement}'

    @classmethod
    def from_message(cls, statement, exc_message):
        m = _match_exc_message(exc_message)
        if m:
            return cls(
                statement=statement,
                code=int(m.group('code')),
                display_text=m.group('display_text'),
                stack_trace=m.group('stack_trace'),
            )
        else: # pragma: nocover
            return cls(
                statement=statement, code=None, display_text=exc_message,
            )



from datetime import date, datetime
from decimal import Decimal
from enum import Enum

import pytest
import sqlalchemy as sa

import aiochsa


async def test_ddl(conn, table_test1):
    await conn.execute(sa.DDL(f'DROP TABLE {table_test1.name}'))


@pytest.mark.xfail(
    raises=TypeError,
    # Is the game worth the candle? For it compilation can't be separated from
    # execution.
    reason='Execution of default is not supported yet',
)
async def test_execute_default(conn, table_test1):
    ts_before = datetime.utcnow().replace(microsecond=0)
    now = await conn.fetchval(table_test1.c.timestamp.default)
    ts_after = datetime.utcnow().replace(microsecond=0)
    assert ts_before <= now <= ts_after


async def test_func(conn):
    ts_before = datetime.utcnow().replace(microsecond=0)
    now = await conn.fetchval(sa.func.now())
    ts_after = datetime.utcnow().replace(microsecond=0)
    assert ts_before <= now <= ts_after


async def test_func_params(conn):
    result = await conn.fetchval(
        sa.func.plus(sa.bindparam('a'), sa.bindparam('b'))
            .params({'a': 12, 'b': 23})
    )
    assert result == 35


async def test_func_params_args(conn):
    result = await conn.fetchval(
        sa.func.plus(sa.bindparam('a'), sa.bindparam('b')),
        {'a': 12, 'b': 23},
    )
    assert result == 35


async def test_simple_round(conn, table_test1):
    now = datetime.utcnow().replace(microsecond=0)
    values = {
        'id': 1,
        'enum': 'ONE',
        'name': 'test',
        'timestamp': now,
        'amount': Decimal('1.23'),
    }
    await conn.execute(
        table_test1.insert()
            .values(values)
    )

    rows = await conn.fetch(table_test1.select())
    row1 = await conn.fetchrow(table_test1.select())
    row2 = await conn.fetchrow(f'SELECT * from {table_test1.name}')
    assert rows == [row1]
    assert row1 == row2 == values


async def test_non_ascii(conn, table_test1):
    sample = 'зразок'
    await conn.execute(
        table_test1.insert()
            .values(id=1, name=sample)
    )
    name = await conn.fetchval(
        sa.select([table_test1.c.name])
            .where(table_test1.c.id == 1)
    )
    assert name == sample


async def test_enum(conn, table_test1):

    class EnumType(str, Enum):
        ONE = 'ONE'
        TWO = 'TWO'

    await conn.execute(
        table_test1.insert()
            .values(id=1, enum=EnumType.ONE)
    )
    value = await conn.fetchval(
        sa.select([table_test1.c.enum])
            .where(table_test1.c.id == 1)
    )
    assert value == EnumType.ONE


async def test_unsupported_type(conn):
    with pytest.raises(TypeError):
        await conn.fetchval(
            sa.select([sa.bindparam('a')])
                .params(a=...)
        )


@pytest.mark.xfail(reason='Feature is disabled in favour of speed')
async def test_defaults(conn, table_test1):
    ts_before = datetime.utcnow().replace(microsecond=0)
    values = {
        'id': 1,
        'enum': 'ONE',
        'name': 'test',
    }
    await conn.execute(
        table_test1.insert()
            .values(values)
    )
    ts_after = datetime.utcnow().replace(microsecond=0)

    row = await conn.fetchrow(table_test1.select())
    for field, value in values.items():
        assert row[field] == value
    assert ts_before <= row['timestamp'] <= ts_after
    assert row['amount'] == Decimal(0)


async def test_insert_multiple(conn, table_test1):
    values = [
        {
            'id': i + 1,
            'name': f'test{i + 1}',
            'enum': 'TWO' if i % 2 else 'ONE',
        }
        for i in range(6)
    ]
    await conn.execute(
        table_test1.insert()
            .values(values)
    )

    rows = await conn.fetch(
        sa.select([table_test1.c.id])
            .where(table_test1.c.enum == 'ONE')
            .order_by(table_test1.c.id.desc())
    )
    assert [item_id for (item_id,) in rows] == [5, 3, 1]


async def test_insert_multiple_args(conn, table_test1):
    values = [
        {'id': i + 1, 'name': f'test{i + 1}'}
        for i in range(3)
    ]
    await conn.execute(
        table_test1.insert(), *values,
    )

    rows = await conn.fetch(
        sa.select([table_test1.c.id, table_test1.c.name])
    )
    assert rows == values


async def test_insert_select(conn, table_test1, table_test2):
    values = [
        {'id': i + 1, 'name': f'test{i + 1}'}
        for i in range(3)
    ]
    await conn.execute(
        table_test1.insert(), *values,
    )

    await conn.execute(
        table_test2.insert()
            .from_select(
                [table_test2.c.num, table_test2.c.title],
                sa.select([table_test1.c.id, table_test1.c.name])
                    .where(table_test1.c.id > 1)
            )
    )
    rows = await conn.fetch(
        table_test2.select()
    )
    assert rows == [('test2', 2), ('test3', 3)]


@pytest.mark.xfail(reason='CTE is not supported by clickhouse-sqlalchemy')
async def test_insert_select_cte(conn, table_test2):
    await conn.execute(
        table_test2.insert(), {'title': 'test', 'num': 1},
    )

    max_num = (
        sa.select([sa.func.max(table_test2.c.num)])
            .cte()
    )

    await conn.execute(
        table_test2.insert()
            .from_select(
                [table_test2.c.num, table_test2.c.title],
                sa.select([max_num, table_test2.c.title])
            )
    )
    rows = await conn.fetch(
        table_test2.select()
    )
    assert rows == [('test', 1), ('test', 2)]


async def test_join(conn, table_test1):
    await conn.execute(
        table_test1.insert(),
        *[
            {'id': i + 1, 'name': f'test{i + 1}'}
            for i in range(3)
        ],
    )

    test_alias = table_test1.alias()
    rows = await conn.fetch (
        sa.select([table_test1.c.id, test_alias.c.id])
            .select_from(
                table_test1.join(
                    test_alias,
                    test_alias.c.id == table_test1.c.id,
                )
            )
    )
    assert {tuple(row) for row in rows} == {(1, 1), (2, 2), (3, 3)}


async def test_iterate(conn, table_test1):
    await conn.execute(
        table_test1.insert(),
        *[
            {'id': i + 1, 'name': f'test{i + 1}'}
            for i in range(3)
        ],
    )

    rows_agen = conn.iterate(
        sa.select([table_test1.c.id])
    )
    assert [item_id async for (item_id,) in rows_agen] == [1, 2, 3]


async def test_fetchval_empty(conn, table_test1):
    value = await conn.fetchval(
        sa.select([table_test1.c.id])
    )
    assert value is None


async def test_select_params(conn, table_test1):
    await conn.execute(
        table_test1.insert(),
        *[
            {'id': i + 1, 'name': f'test{i + 1}'}
            for i in range(3)
        ],
    )

    rows = await conn.fetch(
        sa.select([table_test1.c.id])
            .where(table_test1.c.id >= sa.bindparam('min_id'))
            .params(min_id=2)
    )
    assert [item_id for (item_id,) in rows] == [2, 3]


async def test_select_params_args(conn, table_test1):
    await conn.execute(
        table_test1.insert(),
        *[
            {'id': i + 1, 'name': f'test{i + 1}'}
            for i in range(3)
        ],
    )

    rows = await conn.fetch(
        sa.select([table_test1.c.id])
            .where(table_test1.c.id >= sa.bindparam('min_id')),
        {'min_id': 2},
    )
    assert [item_id for (item_id,) in rows] == [2, 3]


async def test_final_hint(conn, table_test3):
    await conn.execute(
        table_test3.insert(),
        *[{'key': 123, 'value': i} for i in range(10)]
    )
    rows = await conn.fetch(
        table_test3.select()
            .with_hint(table_test3, 'FINAL')
    )
    assert len(rows) == 1
    assert rows[0]['value'] == 45


async def test_aggregate_function(conn, table_test4):
    await conn.execute('INSERT INTO test4 SELECT 1, sumState(123)')
    value = await conn.fetchval(
        sa.select([table_test4.c.value])
    )
    assert isinstance(value, aiochsa.types.AggregateFunction)
    value = await conn.fetchval(
        sa.select([sa.func.sumMerge(table_test4.c.value)])
    )
    assert value == 123


async def test_nested_structures(conn):
    value = await conn.fetchval(
        r"SELECT ("
            r"1, "
            r"(2, NULL, 'a\nb\tc\0d\'', ['a', '\'']),"
            r"1.23, toDecimal32('1.23', 2), toDate('2000-01-01')"
        r")"
    )
    assert value == (
        1,
        (2, None, 'a\nb\tc\0d\'', ['a', '\'']),
        1.23, Decimal('1.23'), date(2000, 1, 1)
    )
